/***********************************************************************
 Name             : sigmoid.java
 Autor            : Marcel Lattmann
                    Eidgenoessisch Technische Hochschule Zuerich (ETHZ)
 Letzte Aenderung : 28.7.1999

 Durch die Messpunkt eine Sigmoid-Kurve legen. Haupt-Methode ist fit, 
 die mit Hilfe des Gauss-Newton-Algorithmus die zwei Parameter a und b
 bestimmt.
 
 Benoetigt folgende Klassen:
   Matrix.class	          Datentyp Matrix und Lineare Algebra Operationen

 History:
************************************************************************/
class sigmoid { 
  // P = Matrix fuer die Datenpunkte.
  //     P.element[i][0] = x-Koordinate
  //     P.element[i][1] = y-Koordinate
  public Matrix P = new Matrix();	

  // Kurvenparameter
  public double a,b;

  // Die gesuchten Werte mit 50 und 95% Hemmung     
  public double ed50, ed95;

  // Die sum(residuen) ist ein Mass fuer die Qualitaet des Fits
  public double sumMin;

  // Anzahl Iterationen
  public int iter;

  // Fuer Testausgaben
  public String res;

  public sigmoid() {
    // Konstruktor
    clear();
  };


  public void clear(){
    P=new Matrix();  // Loesche alle Punkte
    res="";   
  }

  public String checkData(String newln){
    /* Macht eine Vorbearbeitung der Daten. Prueft, ob die Daten
       im Bereich 0..1 oder 0..100 liegen und normiert sie auf 0..1.
       Wenn alles ok, dann Rueckgabe-String ="", sonst Fehlermeldung. */
    // Wenn nicht genuegend Punkte vorhanden sind, dann Abbruch
    int i,j;
    String res = "";
    
    /* Testen, wie viele Daten ueber 1 liegen */
    j=0;
    for (i=0; i<P.zeilen; i++) {
      if (P.element[i][1] > 1.01) {
        j++;
      }
    }

    /* Wenn mind. 3 Werte ueber 1 liegen, dann nehmen wir
       an, dass es sich um Prozentzahlen handelt. */
    if (j>3) {
      for (i=0;i<P.zeilen;i++) {
        P.element[i][1] = P.element[i][1] / 100;
      }
    }

    /* Pruefe, ob die Werte immer noch unter 0 oder ueber 1 sind */ 
    for (i=0;i<P.zeilen;i++) {
      //      if (P.element[i][1] < -0.01) {
      //res= "Data underflow. Please correct input data and try again";
      //return res;
      //}
      //else
      if (P.element[i][1] > 1.01) {
        res= "Data overflow. Please correct input data and try again";
        return res;
      }
    }

    // Der Algorithmus vertraegt keine y-Werte, die ca. 0 sind 
    for (i=0;i<P.zeilen;i++) {
      if (Math.abs(P.element[i][1]) < 0.00001)
        P.element[i][1] = 0.00001;
    }
    return res;
  }

  public void fit(String newln) {
    /* Legt die Kurve nach der Kleinstquadratmethode durch die
       Punkte. Dabei wird die Newton-Methode mit automatischer
       Schrittweite verwendet. */

    
    int i;
    double suma, sumb;
    double step;    /* Variable Schrittweite */

    /* sum(residuen) wird als Abbruchkriterium und zur Steuerung
       der Schrittweite gebraucht. */
    double sum;

    // sumMin ist  minimal bei den Koeffizienten mina, minb  
    double mina=0, minb=0;

    double logx[] = new double[P.zeilen];
    double h[] = new double[P.zeilen];
    // Residuen
    double r[] = new double[P.zeilen];
    // Jacobi Matrix J
    double J[][] = new double[P.zeilen][2];

    // Array mit dem Verlauf der Norm
    double sumVec[] = new double[1000];
    
    // Array nach X-Werten aufsteigend sortieren
    P.sort();

    // log(x) wird oefters benoetigt...
    for(i=0; i<P.zeilen; i++){
      logx[i]=Math.log(Math.abs(P.element[i][0]));
    }

    // Startwert ermitteln

    double ys,smin; 
    double a0=0, a1=10, b0=0, b1=2;
    double astep=(a1-a0)/50;
    double bstep=(b1-b0)/50;   
    smin=10000;
    for (a=a0; a<=a1; a+=astep) {
      for (b=b0; b<=b1; b+=bstep) {
        sum=0;
        for (i=0; i<P.zeilen; i++) {
          ys= 1 / (1+Math.exp(-a-b*Math.log(P.element[i][0])));
          sum = sum + (ys-P.element[i][1])*(ys-P.element[i][1]);
	}
	if (sum < smin) {
          smin=sum; mina = a; minb = b;
        }
      }
    }
    a=mina; b=minb;

    // res= res + " QR: a = " + a + newln +
    //  " QR: b = " + b;         // ist noch korrekt...

    // Iteration mit Gauss-Newton
    step = 0.5;
    iter = 1;
    sumMin = 1000;         // Minimale Norm
    while (iter < 1000) {

      /* sum(residuen) mit Hilfe der aktuellen Approximation der
         Parameter a und b berechnen. */
      sum = 0;
      for (i=0; i<P.zeilen; i++) {

        // Hilfsvektor h
        h[i]=Math.exp(-a-b*logx[i]);
        // Residuum
        r[i]=1 / (1+h[i])-P.element[i][1];
        sum = sum + r[i]*r[i];
      }
      sumVec[iter]=sum;

      // Merke Dir das neue Minimum
      if (sum < sumMin) {
        sumMin = sum;
        mina = a;
        minb = b;
      }
      // Schrittweite steuern
      if (iter % 10 == 0) {

        // Schrittweite verkleinern
        if ((std(iter-5,iter,sumVec) > 0.01) && (step > 0.05))
          step /= 2;
      }
      if (iter % 50 == 0) {
        // Abbruchbedingung testen
        if (span(iter-20,iter,sumVec) < 0.0001*sum)
          break;  

        // Schrittweite vergroessern
        if ((std(iter-5,iter,sumVec) < 0.001) && (step < 1.0))
          step *= 1.5;
      }

      // res = res + " iter: " + iter;

      // Jacobi-Matrix berechnen.
      for (i=0; i<P.zeilen; i++) {
        J[i][0] = h[i] / ((1+h[i])*(1+h[i]));
        J[i][1] = J[i][0] * logx[i];
        //res = res + "J["+i+"][0]="+J[i][0]
        //          + "J["+i+"][1]="+J[i][1] + newln;
      }

      /* Neue Koeffizienten a und b berechnen. Die Matrizen-
         Multiplikation wird aus Performance-Gruenden gleich hier 
         gemacht. */
      suma = 0; sumb = 0;
      for (i=0; i<P.zeilen; i++) {
        suma = suma + J[i][0]*r[i];
        sumb = sumb + J[i][1]*r[i];
      }
      a = a - step * suma;    
      b = b - step * sumb;

      // res= res + "a= " + a + " b = " + b;

      iter++;
    }
    a = mina; b= minb;  
    ed50 = Math.exp(- a/b);
    ed95 = Math.exp((-Math.log(0.05/0.95) - a)/b);
  }

  public double f(double x) {
    // Berechnet den Funktionswert f(x) 
    return (1 / (1+Math.exp(-a-b*Math.log(x))));
  }
  
  public double inverse(double y) {
  // Berechnet die Inverse des Sigmoids an der Stelle y
      return Math.exp((-Math.log((1-y)/y) - a)/b);
  }

  public double Xmin(boolean log) {
    /* Liefert das Minimum des X-Bereichs retour.
       Wenn log true ist, dann wird log10(Xmin) retourgegeben.
       Annahme: X-Werte sind sortiert. */ 
    if (P.zeilen == 0)
      return 0.0;
    double xmin = P.element[0][0];
    if (log)
      return Math.log(Math.abs(xmin))/Math.log(10);
    else
      return xmin;
  }

  public double Xmax(boolean log) {
    /* Liefert das Maximum des X-Bereichs retour.
       Wenn log true ist, dann wird log10(Xmax) retourgegeben.
       Annahme: X-Werte sind sortiert. */ 
    if (P.zeilen == 0)
      return 0.0;
    double xmax = P.element[P.zeilen-1][0];
    if (log)
      return Math.log(Math.abs(xmax))/Math.log(10);
    else
      return xmax;
  }

  public double Ymin(boolean log) {
    if (P.zeilen == 0)
      return 0.0;
    double ymin = P.element[0][1];
    for (int i=0; i<P.zeilen; i++) {
      if (P.element[i][1]<ymin)
        ymin = P.element[i][1];
    }
    if (log)
      return Math.log(Math.abs(ymin))/Math.log(10);
    else
      return ymin;
  }

  public double Ymax(boolean log) {
    if (P.zeilen == 0)
      return 0.0;
    double ymax = P.element[0][1];
    for (int i=0; i<P.zeilen; i++) {
      if (P.element[i][1]>ymax)
        ymax = P.element[i][1];
    }
    if (log)
      return Math.log(Math.abs(ymax))/Math.log(10);
    else
      return ymax;
  }

  private double span(int from, int to, double[] vec) {
    /* Ermittelt im Vektor vec im Bereich von from..to
       die Spannweite der Werte. Die Spannweite ist 
       (Maximum Minimum). */
    int i;
    double min = vec[to];
    double max = vec[to];
    for (i=from; i<to; i++) {
      if (vec[i]<min)
        min = vec[i];
      else if (vec[i]>max)
        max = vec[i];
    }
    return (max - min); 
  }

  private double std(int from, int to, double[] vec) {
    /* Berechnet im Vektor vec im Bereich von from..to
       die Standardabweichung. */ 
    int i;
    double mean=0,sum=0;

    for (i=from; i<=to; i++) 
      mean += vec[i];
    mean = mean / (to - from+1);

    for (i=from; i<=to; i++) 
      sum = sum + (vec[i]-mean) * (vec[i] - mean);
    sum = sum / (to - from +1);

    return Math.sqrt(sum); 
  }
}
